#!/usr/bin/env python3

# pumila: convert images to sound waves
# Copyright (C) 2011 Niels Serup

# This file is part of pumila.

# pumila is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# pumila is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with pumila.  If not, see <http://www.gnu.org/licenses/>.

"""
Command-line parsing
"""

import sys
from optparse import OptionParser, OptionGroup
import os
import re
from . import core
from . import misc
from . import info

info.add_metadata(misc.get_selfdict(__name__))

class _SimplerOptionParser(OptionParser):
    """A simplified OptionParser"""

    def format_description(self, formatter):
        return self.description

    def format_epilog(self, formatter):
        return self.epilog.strip() + '\n'

    def add_option(self, *args, **kwds):
        try: kwds['help'] = kwds['help'].strip()
        except KeyError: pass
        return OptionGroup.add_option(self, *args, **kwds)

def parse_args(cmdargs=None):
    """
    parse_args(cmdargs: [str] = sys.argv[1:]) -> SoundCore

    Base actions on input from the command line. Create a SoundCore object from
    the given options.
    """

    if cmdargs is None:
        cmdargs = sys.argv[1:]

    parser = _SimplerOptionParser(
        prog=info.program.name,
        usage='Usage: %prog [OPTION]... [INPUTFILE]...',
        version=info.program.version_info,
        description=info.program.description,
        epilog='''
Input images must be in a format supported by pygame, i.e. either PNG, JPEG,
GIF, BMP, TGA, PCX, TIF, LBM, PMB, PGM, PPM, or XPM (or a subset). There is
also limited support for OpenRaster files.

If you specify more than one input file, the generated sounds from each file
will be mixed. Extra options, such as gain, can be given as comma-seperated
KEY=VALUE pairs after a slash after the filename, like
`testdir/test1.png/gain=0.7,min=300 Hz,max=10 kHz'. `gain' must be a number
between 0 and 1, and defaults to 1. `min' and `max' denotes the lowest and
highest frequency and understand units. `min' defaults to 220 Hz, and `max'
defaults to 2200 Hz.

''')

    parser.add_option('-o', '--output-file', dest='outputfile',
                      metavar='FILENAME', help='''

the name of the file that sound output should be written to. '-' means standard
out. File format will be derived from file extension, if possible. Defaults to
'out.wav'.

''', default='out.wav')

    parser.add_option('-O', '--no-output-file', dest='outputfile',
                      action='store_false', help='''

do not output sound to a file.

''')

    parser.add_option('-f', '--output-format', dest='outputformat',
                      metavar='FILEFORMAT', type='choice', choices=['wav'],
                      help='''

the format of the sound output. Choose between the default WAVE format 'wav'
(playable by all audio players) and nothing else.

''')
        
    parser.add_option('-y', '--overwrite-outfile', dest='overwrite',
                      action='store_true', default=False, help='''

When writing to a file, overwrite it if it exists.

''')

    parser.add_option('-p', '--play', dest='play',
                      action='store_true', help='''

play the sound after it has been generated.

''', default=False)

    parser.add_option('-P', '--play-at-once', dest='playatonce',
                      action='store_true', help='''

play the sound while it is being generated. This will most likely result in
poor audio, if any, unless your computer is very fast.

''', default=False)

    parser.add_option('-d', '--pixel-duration', dest='pixelduration',
                      metavar='DURATION', help='''

how much time one vertical pixel should last. Second-based units can be
used. If no unit is used, milliseconds are used. Defaults to 10 ms. Is ignored
if --full-duration is given.

''', default=None)

    parser.add_option('-D', '--full-duration', dest='fullduration',
                      metavar='DURATION', help='''

how much time the entire sound should last.

''', default=None)

    parser.add_option('-c', '--channels', dest='channels',
                      metavar='INTEGER', type='int', help='''

the number of channels. Defaults to 1 (mono).

''', default=1)
    
    parser.add_option('-w', '--sample-width', dest='samplewidth',
                      metavar='BITS', type='int', help='''

the size of a sample. Not applicable with the .pml format. Defaults to 16 bits.

''', default=16)
    
    parser.add_option('-r', '--framerate', dest='framerate',
                      metavar='INTEGER', type='int', help='''

the number of samples in one second. Defaults to 44100 (Hz).

''', default=44100)

    parser.add_option('-q', '--quiet', dest='verbose',
                      action='store_false', help='''

do not print info or show a progressbar.

''', default=True)

    parser.add_option('-G', '--no-progressbar', dest= 'showprogressbar',
                  action= 'store_false', help= '''

do not show a progressbar. This option is not needed if --quiet is
specified. Not showing the progressbar can often speed the process up by more
than 100 milliseconds.

''', default=True)

    parser.add_option('-m', '--add-metadata', dest='metadata', nargs=2,
                      action='append', default=[], metavar='KEY VALUE',
                      help='''

add a key-value metadata set. Meaningful KEY strings include "title", "author",
and "description". This option currently does nothing.

''')

    o, args = parser.parse_args(cmdargs)
        
    o.inputfiles = []
    for x in args:
        if os.path.exists(x):
            o.inputfiles.append(x)
        elif '/' in x:
            path, settings = x.rsplit('/', 1)
            settings = {k: v for k, v in (re.split('\s*=\s*', x)
                                          for x in settings.split(','))}
            o.inputfiles.append((path, settings))
            
    if not o.inputfiles:
        parser.print_help()
        print()
        parser.error('no input file specified')

    return core.SoundCore(*o.inputfiles, channels=o.channels,
                           samplewidth=o.samplewidth, framerate=o.framerate,
                           pixelduration=o.pixelduration,
                           fullduration=o.fullduration, play=o.play,
                           playatonce=o.playatonce, outputfile=o.outputfile,
                           outputformat=o.outputformat, overwrite=o.overwrite,
                           metadata=o.metadata, verbose=o.verbose,
                           showprogressbar=o.showprogressbar)



#!/usr/bin/env python3

# pumila: convert images to sound waves
# Copyright (C) 2011 Niels Serup

# This file is part of pumila.

# pumila is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# pumila is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with pumila.  If not, see <http://www.gnu.org/licenses/>.

"""
Loads images.
"""

import sys
import os.path
import tempfile
import zipfile
import re
import pygame
import numpy

def load(path):
    """Load image from path."""
    if path.endswith('.ora'):
        return load_ora(path)
    try:
        surf = pygame.image.load(path)
        try:
            rgb = pygame.surfarray.pixels3d(surf)
        except pygame.error:
            rgb = pygame.surfarray.array3d(surf)
        try:
            alpha = pygame.surfarray.pixels_alpha(surf)
        except (ValueError, pygame.error):
            alpha = numpy.empty((len(rgb), len(rgb[0])), dtype=numpy.uint8)
            alpha[:] = 255
        return (rgb, alpha)
    except pygame.error:
        raise ValueError('file {} is not loadable'.format(repr(path)))

def load_ora(path):
    """Load OpenRaster image from path."""

    temp_path = tempfile.mkdtemp()
    zf = zipfile.ZipFile(path)
    zf.extractall(temp_path)

    with open(os.path.join(temp_path, 'stack.xml')) as f:
        xml = f.read()
        
    files = sorted(filter(lambda x: not x.endswith('background.png'),
                          ('{}/{}'.format(temp_path, x.strip('"\''))
                           for x in re.findall(r'src=(.+?.png)', xml))))
    return tuple(map(load, files))



#!/usr/bin/env python3
# cython: language_level=3

# pumila: convert images to sound waves
# Copyright (C) 2011 Niels Serup

# This file is part of pumila.

# pumila is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# pumila is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with pumila.  If not, see <http://www.gnu.org/licenses/>.

"""
Generates sound.
"""

import itertools
import colorsys
import math
import struct
from . import primitives

cdef double _onefour, _twofour, _threefour
_onefour, _twofour, _threefour = 1./4, 2./4, 3./4

cdef class SoundGenerator:
    """
    The actual resource-demanding sound wave generation. Only works when
    subclassed by SoundCore.
    """

    def get_samples(self):
        cdef int i, j, incr, channels, onepixsamplen
        cdef double sampnum, step, total
        channels, onepixsamplen = self.channels, self.one_pixel_samples_len
        sampnum, step = 0, 1.0 / self.framerate
        imgs_range, freq_range, gains = self.imgs_range, self.freq_range, self.gains
        rgbs, alphas = self.rgbs, self.alphas
        for imgsrgb, imgsalpha in itertools.zip_longest(
            itertools.zip_longest(*rgbs),
            itertools.zip_longest(*alphas)):
            pvars = tuple(tuple(filter(lambda x: x is not None,
                    (self.rgbafg_to_wavefunc(rgb[0], rgb[1], rgb[2], a, freq, gain)
                    for rgb, a, freq in (itertools.zip_longest(
                                    imgsrgb[j], imgsalpha[j], freq_range[j])))))
                          if imgsalpha[j] is not None else None
                          for j, gain in itertools.zip_longest(imgs_range, gains))

            for i in range(onepixsamplen):
                total, incr = 0, 0
                for pvar in pvars:
                    if pvar is None:
                        continue
                    for sampfunc in pvar:
                        total += sampfunc(sampnum)
                        incr += 1
                samp = total / incr if incr > 0 else 0
                for j in range(channels):
                    yield samp
                sampnum += step

    def rgbafg_to_wavefunc(self, int r, int g, int b, int ai,
                           double freq, double gain):
        """
        Convert rgb, alpha, frequency, and gain to a sampnum-to-sample
        function.
        """
           # transparent  # saturation == 0, if r == g == b == 0, value == 0
        if ai == 0 or     r == g == b:
            return None # pixel is either transparent or colorless, and does
                        # not matter.

        cdef double h, s, v, a
        h, s, v = colorsys.rgb_to_hsv(r / 255., g / 255., b / 255.)
        a = ai / 255.
        cdef double amp, ar, br, l1, l2, l3, l4
        amp = float(4 * gain * s * v * a)
        if _onefour > h >= 0:
            ar, br = float(_onefour - h), float(h)
            aw, bw = primitives.sine, primitives.triangle
        elif _twofour > h >= _onefour:
            ar, br = float(_twofour - h), float(h - _onefour)
            aw, bw = primitives.triangle, primitives.square
        elif _threefour > h >= _twofour:
            ar, br = float(_threefour - h), float(h - _twofour)
            aw, bw = primitives.square, primitives.sawtooth
        else:
            ar, br = float(1 - h), float(h - _threefour)
            aw, bw = primitives.sawtooth, primitives.sine
            
        l1, l2, l3, l4 = self.waves[freq]
        def _wf(double s):
            return amp * (aw(s, freq, l1, l2, l3, l4) * ar + bw(s, freq, l1, l2, l3, l4) * br)
        return _wf

    def generate(self):
        cdef int i, j, pbar_interval
        if self.showprogressbar:
            pbar_interval, pbar_update = self.pbar_interval, self.pbar.update
        if self.outputformat == 'wav':
            wavof = self.wavof
            def wavof_write(val):
                wavof.writeframesraw(struct.pack('<h', val))
        if self.play:
            soundarr = self.soundarr
            i, j = self.sound_offset, 0
            if self.outputformat == 'wav':
                if self.showprogressbar:
                    for x in self.get_samples():
                        if j % pbar_interval == 0: pbar_update(j)
                        val = int(x * 32767)
                        soundarr[i] = val
                        wavof_write(val)
                        i += 1; j += 1
                else:
                    for x in self.get_samples():
                        val = int(x * 32767)
                        soundarr[i] = val
                        wavof_write(val)
                        i += 1; j += 1
            else:
                if self.showprogressbar:
                    for x in self.get_samples():
                        if j % pbar_interval == 0: pbar_update(j)
                        soundarr[i] = int(x * 32767)
                        i += 1; j += 1
                else:
                    for x in self.get_samples():
                        soundarr[i] = int(x * 32767)
                        i += 1; j += 1
        elif self.outputformat == 'wav':
            i = 0
            if self.showprogressbar:
                for x in self.get_samples():
                    if i % pbar_interval == 0: pbar_update(i)
                    wavof_write(int(x * 32767))
                    i += 1
            else:
                for x in self.get_samples():
                    wavof_write(int(x * 32767))
                    i += 1
        else:
            if self.showprogressbar:
                i = 0
                for x in self.get_samples():
                    if i % pbar_interval == 0: pbar_update(i)
                    i += 1
            else:
                for x in self.get_samples():
                    pass

    def end(self):
        """Finalize objects."""
        pass



#!/usr/bin/env python3
# cython: language_level=3

# pumila: convert images to sound waves
# Copyright (C) 2011 Niels Serup

# This file is part of pumila.

# pumila is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# pumila is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with pumila.  If not, see <http://www.gnu.org/licenses/>.

"""
Primitive waves.
"""

import math
from fractions import Fraction

cdef double _pi2
_pi2 = 2.0 * math.pi

cpdef double sine(double t, double freq, double quarter_wave_len,
                  double half_wave_len, double three_quarter_wave_len,
                  double wave_len):
    return math.sin(t * _pi2 * freq)

cpdef double triangle(double t, double freq, double quarter_wave_len,
                  double half_wave_len, double three_quarter_wave_len,
                  double wave_len):
    cdef double tr
    tr = t % wave_len
    if tr < quarter_wave_len:
        return tr / quarter_wave_len
    elif tr < half_wave_len:
        return 1 - (tr - quarter_wave_len) / quarter_wave_len
    elif tr < three_quarter_wave_len:
        return -(tr - half_wave_len) / quarter_wave_len
    else:
        return (tr - three_quarter_wave_len) / quarter_wave_len - 1

cpdef double square(double t, double freq, double quarter_wave_len,
                  double half_wave_len, double three_quarter_wave_len,
                  double wave_len):
    return 1 if t % wave_len < half_wave_len else -1

cpdef double sawtooth(double t, double freq, double quarter_wave_len,
                  double half_wave_len, double three_quarter_wave_len,
                  double wave_len):
    cdef double tr
    tr = t % wave_len
    if tr < half_wave_len:
        return tr / half_wave_len
    else:
        return (tr - half_wave_len) / half_wave_len - 1

def getlengths(double freq):
    """Get wave length * 1/4, * 1/2, * 3/4, and * 1 for a frequency."""
    cdef double quarter_wave_len, half_wave_len, three_quarter_wave_len, wave_len
    wave_len = 1 / freq
    half_wave_len = wave_len / 2
    quarter_wave_len = half_wave_len / 2
    three_quarter_wave_len = quarter_wave_len * 3
    return (quarter_wave_len, half_wave_len, three_quarter_wave_len, wave_len)



#!/usr/bin/env python3

# pumila: convert images to sound waves
# Copyright (C) 2011 Niels Serup

# This file is part of pumila.

# pumila is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# pumila is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with pumila.  If not, see <http://www.gnu.org/licenses/>.

"""
Misc. objects.
"""

import sys
import os
import locale
import datetime
import time
import email.utils
import itertools
import functools
import collections
import atexit
import traceback

preferred_encoding = locale.getpreferredencoding()

####################
# Helper functions #
####################
class Container:
    pass

class AttributeDict(dict):
    """A dictionary where x.a == x['a']"""
    def __init__(self, *args, **kwds):
        self._to_apply=kwds.get('to_apply')
        self._apply_what=kwds.get('apply_what')
        if args:
            kwds = {}
            nkwds = self._init(args, {})
            for k, v in nkwds.items():
                kwds[k] = v
        dict.__init__(self, **kwds)

    def _init(self, args, kwds):
        nkwds = AttributeDict()
        for i in range(0, len(args), 2):
            k, v = args[i:i + 2]
            rv = None
            if isinstance(v, list):
                v = self._init(v, kwds)
            else:
                if isinstance(v, str):
                    v = v.format(**kwds)
                if self._to_apply and (self._apply_what is None or
                                       type(v) in self._apply_what):
                    rv = self._to_apply(v)
            if not rv: rv = v
            kwds[k] = v
            nkwds[k] = rv
        return nkwds

    __getattr__ = lambda self, k: self.__getitem__(k)
    __setattribute__ = lambda self, k, v: self.__setitem__(k, v) \
        if not k.startswith('_') else None
    __delattr__ = lambda self, k: tryorpass(KeyError, self.__delitem__, k)

    def __setstate__(self, adict):
        for k, v in adict.items():
            self.__setitem__(k, v)

class PeriodTextTuple(tuple):
    """A tuple meant for version numbers"""
    def __init__(self, *args):
        self.text = '.'.join(str(x) for x in self)

    __new__ = lambda self, *xs: tuple.__new__(self, xs)
    __str__ = lambda self: self.text

def tryorpass(exceptions, func=None, *args, **kwds):
    """
    Try to execute a function. Pass if an exception of exceptions is raised.

    If func returns an iterable, all elements will be tested individually. This
    function can be used as a decorator.
    """
    if not func: # Act as decorator
        nfunc = lambda func: tryorpass(exceptions, func, *args, **kwds)
        kwds['return_func'] = nfunc
        return nfunc
    return_func = kwds.get('return_func')
    if return_func:
        del kwds['return_func']
        return_func.__name__ = func.__name__
        return_func.__doc__ = func.__doc__
    _return = lambda r=None: r if not return_func else return_func
    try:
        ret = func(*args, **kwds)
    except exceptions:
        return _return()
    try:
        rets = iter(ret)
    except TypeError:
        return _return(ret)
    rets_list = []
    while True:
        try:
            temp = next(rets)
        except StopIteration:
            break
        except exceptions:
            continue
        rets_list.append(temp)
    return _return(rets_list)

def is_acceptable_file(path):
    """
    Return True if path is file and does not end with '~' or '.py[co]?', else
    False.
    """
    if not os.path.isfile(path):
        return False
    for x in ('~', '.py', '.pyc', '.pyo'):
        if path.endswith(x):
            return False
    return True

def callable(obj):
    """Return True if obj is callable, else False."""
    return isinstance(obj, collections.Callable)

def get_selfdict(name):
    """Get the globals of a module"""
    return sys.modules[name].__dict__

donothing = lambda *x, **y: None

###########
# Logging #
###########

def dateformat(dt=None, localtime=True):
    """Format a date according to RFC 2822"""
    return email.utils.formatdate(
        timeval=time.mktime(dt.timetuple()) if dt else None,
        localtime=localtime)

_logfile, _logfileclose = sys.stderr, lambda: _logfile.close()
_altlogfile, _altlogfileclose = None, lambda: _altlogfile.close()
_log_levels = ('notice', 'warning', 'error')
def log(*message, **kwds):
    """Write a message to the log"""
    level, module, trace = kwds.get('level'), kwds.get('module'), \
        kwds.get('traceback')
    if level is None:
        if kwds.get('error'):
            level = 2
        elif kwds.get('warning'):
            level = 1
        else:
            level = 0
    for k in ('level', 'module', 'error', 'warning', 'traceback'):
        try: del kwds[k]
        except KeyError: pass
    sep = kwds.get('sep', ' ')
    if _logfile or _altlogfile:
        text = '[{}] [{}] {}{}'.format(
            dateformat(), _log_levels[level], '({}) '.format(
                module) if module else '', sep.join(map(str, message)))
        if trace:
            text += '\n' + format(traceback.format_exc())
        for o in (_logfile, _altlogfile):
            if o:
                kwds['file'] = o
                print(text, **kwds)

def newlog(module_name=''):
    """Create a new log function with a module's name"""
    return functools.partial(log, module=module_name)

def setlogfile(fobj=None, close_on_exit=None, alt=False):
    """Set the output file of the log file"""
    global _logfile, _altlogfile
    if fobj is None:
        fobj = sys.stderr
    if close_on_exit is None:
        if fobj in (sys.stderr, sys.stdout):
            close_on_exit = False
    else:
        close_on_exit = True

    if isinstance(fobj, str):
        try:
            if not alt:
                _logfile = open(fobj, 'w')
            else:
                _altlogfile = open(fobj, 'w')
        except Exception:
            pass
    else:
        if not alt:
            _logfile = fobj
        else:
            _altlogfile = fobj
    lclose = _logfileclose if not alt else _altlogfileclose
    if close_on_exit:
        atexit.register(lclose)
    else:
        atexit.unregister(lclose)

def setnologfile():
    """Do not log"""
    global _logfile
    _logfile = None
    atexit.unregister(_logfileclose)

def setaltlogfile(fobj, close_on_exit=None):
    """Set the alternative output file of the log file"""
    setlogfile(fobj, close_on_exit, alt=True)




#!/usr/bin/env python3

# pumila: convert images to sound waves
# Copyright (C) 2011 Niels Serup

# This file is part of pumila.

# pumila is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# pumila is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with pumila.  If not, see <http://www.gnu.org/licenses/>.

"""
Core information about program
"""

import os

from . import misc

_selfdict, log = misc.get_selfdict(__name__), misc.newlog('info')


program = misc.AttributeDict(
    'name',               'pumila',
    'version',            misc.PeriodTextTuple(0, 1, 0),
    'description',        'convert images to sound waves',
    'author',             'Niels Serup',
    'author_email',       'ns@metanohi.name',
    'author_with_email',  '{author} <{author_email}>',
    'url',                'http://metanohi.name/projects/pumila/',
    'copyright',          'Copyright (C) 2011  Niels Serup',
    'short_license_name', 'AGPLv3+',
    'short_license',      '''\
License AGPLv3+: GNU AGPL version 3 or later <http://gnu.org/licenses/agpl.html>
This is free software: you are free to change and redistribute it.
There is NO WARRANTY, to the extent permitted by law.''',
    'version_info',       '{name} {version.text}\n{copyright}\n{short_license}',
    )

home_dir = os.path.expanduser('~')
localpaths = misc.AttributeDict(
    'dirs', [
        'root',        '.pumila',
        ],
    'files', [
        'logfile',     '{root}/.log',
        ],
    to_apply=lambda p: os.path.normcase(os.path.normpath(
            os.path.join(home_dir, p))),
    apply_what=(str,)
    )

@misc.tryorpass(Exception)
def createlocaldirs():
    for v in localpaths.dirs.values():
        os.makedirs(v)
        yield

misc.setaltlogfile(localpaths.files.logfile)

def add_metadata(globdict):
    globdict['__version__'] = program.version.text
    globdict['__author__'] = program.author_with_email

add_metadata(_selfdict)



#!/usr/bin/env python3

# pumila: convert images to sound waves
# Copyright (C) 2011 Niels Serup

# This file is part of pumila.

# pumila is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# pumila is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with pumila.  If not, see <http://www.gnu.org/licenses/>.

"""
Sound-related units.
"""

from fractions import Fraction
from . import misc
from . import info

info.add_metadata(misc.get_selfdict(__name__))

_prefixes = {k: Fraction(10)**v for k, v in {
    'y': -24,
    'z': -21,
    'a': -18,
    'f': -15,
    'p': -12,
    'n': -9,
    'μ': -6,
    'u': -6,
    'm': -3,
    'k': 3,
    'M': 6,
    'G': 9,
    'T': 12,
    'P': 15,
    'E': 18,
    'Z': 21,
    'Y': 24,
    }.items()}

class UnitError(ValueError):
    pass

class NumberWithUnit:
    def __init__(self, unit_name, number=1):
        if not unit_name:
            raise UnitError
        if len(unit_name) > 1 and unit_name[0] in _prefixes.keys():
            self.unit = unit_name[1:]
            self.base_number = number * _prefixes[unit_name[0]]
        else:
            self.unit = unit_name
            self.base_number = number
        self.number = number

    def get_number(self, unit_name=None):
        if not unit_name or unit_name == self.unit:
            return self.number
        if not unit_name.endswith(self.unit) or \
                (len(unit_name) > 1 and unit_name[0] not in _prefixes.keys()):
            raise UnitError('unit {} does not exist'.format(unit_name))
        return self.base_number / _prefixes[unit_name[0]]



#!/usr/bin/env python3

# pumila: convert images to sound waves
# Copyright (C) 2011 Niels Serup

# This file is part of pumila.

# pumila is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# pumila is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with pumila.  If not, see <http://www.gnu.org/licenses/>.

"""
Wrapper for sound generation.
"""

import sys
import os
from fractions import Fraction
import itertools
import re
import numpy
import pygame
import colorsys
import time
import collections
import math
import wave
try:
    import progressbar
except ImportError:
    progressbar = None

from .generate import SoundGenerator
from . import units
from . import image
from . import primitives
from . import misc
from . import info

_selfdict, log = misc.get_selfdict(__name__), misc.newlog('core')
info.add_metadata(_selfdict)

class SoundCore(SoundGenerator):
    def __init__(self, *inputfiles, channels=1, samplewidth=16, framerate=44100,
                 pixelduration=None, fullduration=None, play=False,
                 playatonce=False, outputfile=None, outputformat=None,
                 overwrite=False, metadata={}, returndata=None, verbose=False,
                 showprogressbar=False):
        """
        Generate sound waves.

        If returndata, return a list of numbers.
        """
        self.inputfiles = []
        for path in inputfiles:
            if isinstance(path, tuple) or isinstance(path, list):
                path, settings = path[0], self.parse_settings(path[1])
            else:
                settings = {}
            if not os.path.isfile(path):
                raise OSError('{} does not exist'.format(repr(path)))
            if not 'gain' in settings:
                settings['gain'] = 1
            if not 'min' in settings:
                settings['min'] = 220
            if not 'max' in settings:
                settings['max'] = 2200
            for x in ('gain', 'min', 'max'):
                if settings[x].denominator == 1:
                    settings[x] = settings[x].numerator
            self.inputfiles.append((path, settings))

        self.channels, self.samplewidth, self.framerate, self.pixelduration, \
            self.fullduration, self.play, self.playatonce, self.outputfile, \
            self.outputformat, self.overwrite, self.metadata, self.returndata, \
            self.verbose, self.showprogressbar = \
            channels, samplewidth, framerate, pixelduration, fullduration, \
            play, playatonce, outputfile, outputformat, overwrite, metadata, \
            returndata, verbose, showprogressbar

        if self.playatonce:
            self.play = True
        elif not outputfile and outputformat:
            self.outputformat = None
        elif outputfile:
            if outputformat:
                self.outputformat = outputformat.lower()
            elif outputfile.lower().endswith('.pml'):
                self.outputformat = 'pml'
            else:
                self.outputformat = 'wav' # Default
            if self.outputformat not in ('pml', 'wav'):
                raise ValueError('{} is not an accepted format'.format(
                        self.outputformat))
            if outputfile == '-':
                self.outputfile = open(1, 'wb')
            elif os.path.isfile(outputfile) and not overwrite:
                raise ValueError('file {} already exists'.format(
                        repr(outputfile)))

        if self.pixelduration:
            self.pixelduration = self._unit_parse(self.pixelduration, 'ms')
        if self.fullduration:
            self.fullduration = self._unit_parse(self.fullduration, 'ms')
        elif not self.pixelduration:
            self.pixelduration = 10 # ms, default

        self.log = log if self.verbose else misc.donothing
        if not self.verbose or not progressbar:
            self.showprogressbar = False

    def parse_settings(self, settings):
        """
        parse_settings(settings: dict) -> dict

        Remove pairs where key is not in ('gain', 'min', 'max') and parse units.
        """
        # Warning: a new dict is not created
        for k, v in settings.items():
            if k not in ('gain', 'min', 'max'):
                del settings[k]
            elif k in ('gain', 'min', 'max'):
                val = self._unit_parse(v, 'Hz')
                if val:
                    settings[k] = val
                else:
                    del settings[k]
        return settings

    def _unit_parse(self, val, default_unit=None):
        try:
            return Fraction(val)
        except ValueError:
            try:
                m = re.match(r'([0-9./]+)\s*(.+)\s*$', val)
            except TypeError:
                return
            if m:
                number, unit = Fraction(m.group(1)), m.group(2)
                wn = units.NumberWithUnit(unit, number)
                if default_unit:
                    return wn.get_number(default_unit)
                else:
                    return wn

    def run(self):
        """Generate the sound waves."""
        self.log('Loading images...')
        self.indata = []
        for path, sett in self.inputfiles:
            loaded = image.load(path)
            if isinstance(loaded[0], tuple):
                for img in loaded:
                    self.indata.append((img, sett))
            else:
                self.indata.append((loaded, sett))
        self.log('Loaded {} images.'.format(len(self.indata)))

        maxlen = max(len(x[0][1]) for x in self.indata)
        if not self.pixelduration:
            self.pixelduration = Fraction(self.fullduration, maxlen)
        elif not self.fullduration:
            self.fullduration = self.pixelduration * maxlen
        self.log('Pixel duration wants to be {} ms, full duration wants to be {} ms.'.format(
                self.pixelduration, self.fullduration))

        self.duration_ratio = self.fullduration / self.pixelduration
        self.one_pixel_samples_len = int(self.pixelduration *
                                         self.framerate / 1000)
        self.t_samples_len = int(self.one_pixel_samples_len * self.duration_ratio)
        self.samples_len = self.t_samples_len * self.channels
        self.duration_ratio = self.fullduration / self.pixelduration
        self.log('Number of samples for one pixel is {}, total number of samples is {}.'.format(
                self.one_pixel_samples_len, self.t_samples_len))

        self.pixelduration = Fraction(1000 * self.one_pixel_samples_len,
                                      self.framerate * self.channels)
        self.fullduration = Fraction(1000 * self.samples_len,
                                     self.framerate * self.channels)
        self.log('Pixel duration is {} ms, full duration is {} ms.'.format(
                self.pixelduration.numerator if self.pixelduration.denominator == 1
                else '~{:.2f}'.format(float(self.pixelduration)), 
                self.fullduration.numerator if self.fullduration.denominator == 1
                else '~{:.2f}'.format(float(self.fullduration))))

        data, settings = itertools.zip_longest(*self.indata)
        self.rgbs = tuple(x[0] for x in data)
        self.alphas = tuple(x[1] for x in data)
        self.gains = tuple(float(x['gain']) for x in settings)
        self.imgs_range = tuple(range(len(settings)))

        self.waves = {}
        self.freq_range = []
        i = 0
        for x in settings:
            row_height = len(self.alphas[i][0])
            diff = x['max'] - x['min']
            ratio = diff / row_height
            freqs = tuple(float(ratio * r + x['min']) for r in reversed(
                    range(row_height)))
            self.freq_range.append(freqs)
            for freq in freqs:
                self.waves[freq] = primitives.getlengths(freq)
            i += 1

        if self.returndata:
            return self.get_samples()
        if self.outputformat == 'wav':
            self.wavof = wave.open(self.outputfile, 'w')
            self.wavof.setnchannels(self.channels)
            self.wavof.setsampwidth(self.samplewidth // 8)
            self.wavof.setframerate(self.framerate)

        if self.play:
            pygame.mixer.init(self.framerate, -16, self.channels, 1024)
            self.sound_offset = int(self.channels * self.framerate / 5) \
                if self.playatonce else 0
            soundarr = numpy.empty(self.samples_len + self.sound_offset,
                                   dtype=numpy.int16)
            sound = pygame.sndarray.make_sound(soundarr)
            self.soundarr = pygame.sndarray.samples(sound)
            if self.playatonce:
                self.log('Starting playback...')
                sound.play()
                play_start = time.time()

        self.log('Generating sound...')
        gen_start = time.time()
        if self.showprogressbar:
            self.pbar = progressbar.ProgressBar(maxval=self.samples_len).start()
             # few have a terminal with a width above 300
            self.pbar_interval = self.samples_len // 300
        self.generate()
        if self.showprogressbar:
            self.pbar.finish()
        gen_diff = time.time() - gen_start
        self.log('Sound has been generated. The process took {:.1f} seconds{}.'.format(
                gen_diff, " (that's more than {} minutes!)".format(int(gen_diff // 60))
                if gen_diff / 60 > 3 else ''))

        if self.play:
            if self.playatonce:
                diff = time.time() - play_start
                wait = math.ceil(self.fullduration - diff * 1000 +
                                 1000 * self.sound_offset / self.framerate)
            else:
                self.log('Starting playback...')
                sound.play()
                wait = math.ceil(self.fullduration)
            pygame.time.wait(wait)

    def end(self):
        """Finalize objects."""
        SoundGenerator.end(self)
        if self.outputformat == 'wav':
            self.wavof.close()
        if self.play:
            pygame.mixer.quit()



#!/usr/bin/env python3

from distutils.core import setup
from distutils.extension import Extension
from Cython.Distutils import build_ext

from pumila.info import program as p

cython_files = ('generate', 'primitives')
ext_modules = [Extension('pumila.{}'.format(f), ['pumila/{}.pyx'.format(f)])
               for f in cython_files]

setup(
    name=p.name,
    version=p.version.text,
    author=p.author,
    author_email=p.author_email,
    url=p.url,
    description=p.description,
    long_description=open('README.txt').read(),
    license=p.short_license_name,
    packages=['pumila'],
    scripts=['scripts/pumila', 'scripts/pumila-show'],
    cmdclass={'build_ext': build_ext},
    ext_modules=ext_modules,
    classifiers=['Development Status :: 3 - Alpha',
                 'Intended Audience :: Developers',
                 'Intended Audience :: End Users/Desktop',
                 'Intended Audience :: Science/Research',
                 'License :: OSI Approved :: GNU Affero General Public License v3',
                 'License :: DFSG approved',
                 'Operating System :: OS Independent',
                 'Programming Language :: Python :: 3.1',
                 'Programming Language :: Cython',
                 'Environment :: Console',
                 'Topic :: Utilities',
                 'Topic :: Multimedia :: Sound/Audio :: Sound Synthesis',
                 ])



#!/usr/bin/env python3

# pumila: convert images to sound waves
# Copyright (C) 2011 Niels Serup

# This file is part of pumila.

# pumila is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# pumila is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with pumila.  If not, see <http://www.gnu.org/licenses/>.

"""
pumila is an image-to-sound converter. It transforms variations in color,
brightness, saturation and opacity into different sounds.
"""
# See README.txt for more information.

from . import info

__version__ = info.program.version.text
__author__ = info.program.author_with_email
__copyright__ = '''
# pumila: convert images to sound waves
# Copyright (C) 2011 Niels Serup

# This file is part of pumila.

# pumila is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# pumila is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with pumila.  If not, see <http://www.gnu.org/licenses/>.
'''
